==> osmo-s1gw_0.3.0.184.f880.202507272026.tar.xz: BUILD/src/sctp_proxy.erl <==
%% Copyright (C) 2024 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
%% Author: Vadim Yanitskiy <vyanitskiy@sysmocom.de>
%%
%% All Rights Reserved
%%
%% SPDX-License-Identifier: AGPL-3.0-or-later
%%
%% This program is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Affero General Public License as
%% published by the Free Software Foundation; either version 3 of the
%% License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU Affero General Public License
%% along with this program.  If not, see <https://www.gnu.org/licenses/>.
%%
%% Additional Permission under GNU AGPL version 3 section 7:
%%
%% If you modify this Program, or any covered work, by linking or
%% combining it with runtime libraries of Erlang/OTP as released by
%% Ericsson on https://www.erlang.org (or a modified version of these
%% libraries), containing parts covered by the terms of the Erlang Public
%% License (https://www.erlang.org/EPLICENSE), the licensors of this
%% Program grant you additional permission to convey the resulting work
%% without the need to license the runtime libraries of Erlang/OTP under
%% the GNU Affero General Public License. Corresponding Source for a
%% non-source form of such a combination shall include the source code
%% for the parts of the runtime libraries of Erlang/OTP used as well as
%% that of the covered work.

-module(sctp_proxy).
-behaviour(gen_statem).
-behaviour(sctp_client).

-export([init/1,
         callback_mode/0,
         connecting/3,
         connected/3,
         code_change/4,
         terminate/3]).
-export([start_link/2,
         send_data/2,
         fetch_info/1,
         shutdown/1]).

-include_lib("kernel/include/logger.hrl").
-include_lib("kernel/include/inet.hrl").
-include_lib("kernel/include/inet_sctp.hrl").

-include("s1gw_metrics.hrl").

%% ------------------------------------------------------------------
%% public API
%% ------------------------------------------------------------------

-spec start_link(ConnInfo, Priv) -> Result
    when ConnInfo :: sctp_server:conn_info(),
         Priv :: sctp_client:cfg(),
         Result :: gen_statem:start_ret().
start_link(ConnInfo, Priv) ->
    gen_statem:start_link(?MODULE, [ConnInfo, Priv], []).


-spec send_data(pid(), binary()) -> ok.
send_data(Pid, Data) ->
    gen_statem:cast(Pid, {send_data, Data}).


-spec fetch_info(pid()) -> proplists:proplist().
fetch_info(Pid) ->
    gen_statem:call(Pid, ?FUNCTION_NAME).


-spec shutdown(pid()) -> ok.
shutdown(Pid) ->
    gen_statem:stop(Pid).


%% ------------------------------------------------------------------
%% gen_statem API
%% ------------------------------------------------------------------

init([ConnInfo, ConnCfg]) ->
    {ok, Pid} = s1ap_proxy:start_link(ConnInfo),
    {ok, connecting,
     #{enb_aid => maps:get(aid, ConnInfo),
       conn_cfg => ConnCfg,
       tx_queue => [],
       handler => Pid}}.


callback_mode() ->
    [state_functions, state_enter].


%% CONNECTING state
connecting(enter, OldState,
           #{conn_cfg := ConnCfg} = S) ->
    ?LOG_INFO("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
    %% Initiate connection establishment with the MME
    {ok, Sock} = sctp_client:connect(ConnCfg),
    {next_state, connecting, S#{sock => Sock},
     [{state_timeout, 2_000, conn_est_timeout}]};

%% Handle connection establishment timeout
connecting(state_timeout, conn_est_timeout, _S) ->
    {stop, {shutdown, conn_est_timeout}};

%% Handle an eNB -> MME data forwarding request (queue)
connecting(cast, {send_data, Data},
           #{tx_queue := Pending} = S) ->
    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_UPLINK_PACKETS_QUEUED),
    s1gw_metrics:gauge_inc(?S1GW_GAUGE_S1AP_PROXY_UPLINK_PACKETS_QUEUED),
    {keep_state, S#{tx_queue := [Data | Pending]}};

%% Handle an #sctp_assoc_change event (connection state)
connecting(info, {sctp, _Socket, MmeAddr, MmePort,
                  {[], #sctp_assoc_change{state = ConnState,
                                          assoc_id = Aid}}}, S) ->
    case ConnState of
        comm_up ->
            ?LOG_NOTICE("MME connection (id=~p, ~p:~p) established",
                        [Aid, MmeAddr, MmePort]),
            {next_state, connected, S#{mme_aid => Aid}};
        _ ->
            ?LOG_NOTICE("MME connection establishment failed: ~p", [ConnState]),
            {stop, {shutdown, conn_est_fail}}
    end;

%% Catch-all for other kinds of SCTP events
connecting(info, {sctp, _Socket, MmeAddr, MmePort,
                  {AncData, Data}}, S) ->
    ?LOG_DEBUG("Unhandled SCTP event (~p:~p): ~p, ~p",
               [MmeAddr, MmePort, AncData, Data]),
    {keep_state, S};

connecting(Event, EventData, S) ->
    handle_event(?FUNCTION_NAME, Event, EventData, S).


%% CONNECTED state
connected(enter, OldState, S0) ->
    ?LOG_INFO("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
    %% Send pending eNB -> MME messages (if any)
    S1 = sctp_send_pending(S0),
    {keep_state, S1};

%% Handle an eNB -> MME data forwarding request (forward)
connected(cast, {send_data, Data}, S0) ->
    sctp_send(Data, S0),
    {keep_state, S0};

%% Handle an #sctp_assoc_change event (MME connection state)
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {[], #sctp_assoc_change{state = ConnState,
                                         assoc_id = Aid}}}, S) ->
    case ConnState of
        comm_up ->
            ?LOG_NOTICE("MME connection (id=~p, ~p:~p) is already established?!?",
                        [Aid, MmeAddr, MmePort]),
            {keep_state, S};
        _ ->
            ?LOG_NOTICE("MME connection state: ~p", [ConnState]),
            {stop, {shutdown, conn_fail}}
    end;

%% Handle an #sctp_sndrcvinfo event (MME -> eNB data)
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {[#sctp_sndrcvinfo{assoc_id = Aid,
                                    stream = SID,
                                    ssn = SSN,
                                    tsn = TSN}], Data}},
          #{sock := Sock,
            enb_aid := EnbAid,
            mme_aid := Aid,
            handler := Pid} = S) ->
    ?LOG_DEBUG("MME connection (id=~p, ~p:~p) -> eNB: ~p",
               [Aid, MmeAddr, MmePort,
                #{tsn => TSN, sid => SID, ssn => SSN,
                  len => byte_size(Data), data => Data}]),
    case s1ap_proxy:process_pdu(Pid, Data) of
        {forward, FwdData} ->
            sctp_server:send_data(EnbAid, FwdData);
        {reply, ReData} ->
            ok = sctp_common:send_data({Sock, Aid}, ReData)
    end,
    {keep_state, S};

%% Catch-all for other kinds of SCTP events
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {AncData, Data}}, S) ->
    ?LOG_DEBUG("Unhandled SCTP event (~p:~p): ~p, ~p",
               [MmeAddr, MmePort, AncData, Data]),
    {keep_state, S};

connected(Event, EventData, S) ->
    handle_event(?FUNCTION_NAME, Event, EventData, S).


%% Event handler for all states
handle_event(State, {call, From}, fetch_info,
             #{handler := Pid} = S0) ->
    ERABs = s1ap_proxy:fetch_erab_list(Pid),
    S1 = maps:with([enb_aid, mme_aid, mme_addr, mme_port], S0),
    S2 = S1#{state => State, erab_list => ERABs},
    Reply = proplists:from_map(S2),
    {keep_state_and_data, {reply, From, Reply}};

handle_event(State, {call, From}, EventData, _S) ->
    ?LOG_ERROR("Unexpected call in state ~p: ~p", [State, EventData]),
    {keep_state_and_data, {reply, From, {error, {unexpected_call, State}}}};

handle_event(State, Event, EventData, _S) ->
    ?LOG_ERROR("Unexpected event ~p in state ~p: ~p", [Event, State, EventData]),
    keep_state_and_data.


code_change(_Vsn, State, S, _Extra) ->
    {ok, State, S}.


terminate(Reason, State, #{handler := Pid} = S) ->
    ?LOG_NOTICE("Terminating in state ~p, reason ~p", [State, Reason]),
    s1ap_proxy:shutdown(Pid),
    case S of
        #{sock := Sock,
          mme_aid := Aid} ->
            sctp_common:shutdown({Sock, Aid}),
            gen_sctp:close(Sock);
        #{sock := Sock} ->
            gen_sctp:close(Sock);
        _ -> ok
    end.


%% ------------------------------------------------------------------
%% private API
%% ------------------------------------------------------------------

%% Send a single message to the MME
sctp_send(Data,
          #{sock := Sock,
            enb_aid := EnbAid,
            mme_aid := Aid,
            handler := Pid}) ->
    case s1ap_proxy:process_pdu(Pid, Data) of
        {forward, FwdData} ->
            ok = sctp_common:send_data({Sock, Aid}, FwdData);
        {reply, ReData} ->
            sctp_server:send_data(EnbAid, ReData)
    end.


%% Send pending messages to the MME
sctp_send_pending(#{tx_queue := Pending} = S) ->
    sctp_send_pending(lists:reverse(Pending), S).

sctp_send_pending([Data | Pending], S) ->
    sctp_send(Data, S),
    s1gw_metrics:gauge_dec(?S1GW_GAUGE_S1AP_PROXY_UPLINK_PACKETS_QUEUED),
    sctp_send_pending(Pending, S);

sctp_send_pending([], S) ->
    S#{tx_queue := []}.

%% vim:set ts=4 sw=4 et:

==> osmo-s1gw_0.4.0.259.0f54.202512091800.tar.xz: BUILD/src/sctp_proxy.erl <==
%% Copyright (C) 2024 by sysmocom - s.f.m.c. GmbH <info@sysmocom.de>
%% Author: Vadim Yanitskiy <vyanitskiy@sysmocom.de>
%%
%% All Rights Reserved
%%
%% SPDX-License-Identifier: AGPL-3.0-or-later
%%
%% This program is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Affero General Public License as
%% published by the Free Software Foundation; either version 3 of the
%% License, or (at your option) any later version.
%%
%% This program is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%% GNU General Public License for more details.
%%
%% You should have received a copy of the GNU Affero General Public License
%% along with this program.  If not, see <https://www.gnu.org/licenses/>.
%%
%% Additional Permission under GNU AGPL version 3 section 7:
%%
%% If you modify this Program, or any covered work, by linking or
%% combining it with runtime libraries of Erlang/OTP as released by
%% Ericsson on https://www.erlang.org (or a modified version of these
%% libraries), containing parts covered by the terms of the Erlang Public
%% License (https://www.erlang.org/EPLICENSE), the licensors of this
%% Program grant you additional permission to convey the resulting work
%% without the need to license the runtime libraries of Erlang/OTP under
%% the GNU Affero General Public License. Corresponding Source for a
%% non-source form of such a combination shall include the source code
%% for the parts of the runtime libraries of Erlang/OTP used as well as
%% that of the covered work.

-module(sctp_proxy).
-behaviour(gen_statem).
-behaviour(sctp_client).

-export([init/1,
         callback_mode/0,
         connecting/3,
         connected/3,
         code_change/4,
         terminate/3]).
-export([start_link/2,
         send_data/2,
         fetch_info/1,
         shutdown/1]).

-include_lib("kernel/include/logger.hrl").
-include_lib("kernel/include/inet.hrl").
-include_lib("kernel/include/inet_sctp.hrl").

-include("s1gw_metrics.hrl").


-type conn_info() :: #{state := atom(),
                       handler := pid(),
                       mme_addr := inet:ip_address(),
                       mme_port := inet:port_number(),
                       enb_aid := gen_sctp:assoc_id(),
                       mme_aid => gen_sctp:assoc_id()
                      }.

-export_type([conn_info/0]).


%% ------------------------------------------------------------------
%% public API
%% ------------------------------------------------------------------

-spec start_link(ConnInfo, Priv) -> Result
    when ConnInfo :: sctp_server:conn_info(),
         Priv :: sctp_client:cfg(),
         Result :: gen_statem:start_ret().
start_link(ConnInfo, Priv) ->
    gen_statem:start_link(?MODULE, [ConnInfo, Priv], []).


-spec send_data(pid(), binary()) -> ok.
send_data(Pid, Data) ->
    gen_statem:cast(Pid, {send_data, Data}).


-spec fetch_info(pid()) -> conn_info().
fetch_info(Pid) ->
    gen_statem:call(Pid, ?FUNCTION_NAME).


-spec shutdown(pid()) -> ok.
shutdown(Pid) ->
    gen_statem:stop(Pid).


%% ------------------------------------------------------------------
%% gen_statem API
%% ------------------------------------------------------------------

init([EnbConnInfo, MmeConnCfg]) ->
    {ok, EnbHandle} = enb_registry:enb_register(),
    {ok, Pid} = s1ap_proxy:start_link(EnbHandle, EnbConnInfo),
    {ok, connecting,
     #{enb_aid => maps:get(aid, EnbConnInfo),
       enb_conn_info => EnbConnInfo,
       mme_conn_cfg => MmeConnCfg,
       tx_queue => [],
       enb_handle => EnbHandle,
       handler => Pid}}.


callback_mode() ->
    [state_functions, state_enter].


%% CONNECTING state
connecting(enter, OldState,
           #{enb_conn_info := EnbConnInfo,
             mme_conn_cfg := MmeConnCfg,
             enb_handle := Handle} = S) ->
    ?LOG_INFO("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
    ok = enb_registry:enb_event(Handle, {?FUNCTION_NAME, EnbConnInfo}),
    %% Initiate connection establishment with the MME
    {ok, Sock} = sctp_client:connect(MmeConnCfg),
    {next_state, connecting, S#{sock => Sock},
     [{state_timeout, 2_000, conn_est_timeout}]};

%% Handle connection establishment timeout
connecting(state_timeout, conn_est_timeout, _S) ->
    {stop, {shutdown, conn_est_timeout}};

%% Handle an eNB -> MME data forwarding request (queue)
connecting(cast, {send_data, Data},
           #{tx_queue := Pending} = S) ->
    s1gw_metrics:ctr_inc(?S1GW_CTR_S1AP_PROXY_UPLINK_PACKETS_QUEUED),
    s1gw_metrics:gauge_inc(?S1GW_GAUGE_S1AP_PROXY_UPLINK_PACKETS_QUEUED),
    {keep_state, S#{tx_queue => [Data | Pending]}};

%% Handle an #sctp_assoc_change event (connection state)
connecting(info, {sctp, _Socket, MmeAddr, MmePort,
                  {[], #sctp_assoc_change{state = ConnState,
                                          assoc_id = Aid}}}, S) ->
    case ConnState of
        comm_up ->
            ?LOG_NOTICE("MME connection (id=~p, ~p:~p) established",
                        [Aid, MmeAddr, MmePort]),
            {next_state, connected, S#{mme_aid => Aid}};
        _ ->
            ?LOG_NOTICE("MME connection establishment failed: ~p", [ConnState]),
            {stop, {shutdown, conn_est_fail}}
    end;

connecting(Event, EventData, S) ->
    handle_event(?FUNCTION_NAME, Event, EventData, S).


%% CONNECTED state
connected(enter, OldState,
          #{enb_handle := Handle} = S0) ->
    ?LOG_INFO("State change: ~p -> ~p", [OldState, ?FUNCTION_NAME]),
    MmeConnInfo = conn_info(?FUNCTION_NAME, S0),
    ok = enb_registry:enb_event(Handle, {?FUNCTION_NAME, MmeConnInfo}),
    %% Send pending eNB -> MME messages (if any)
    S1 = sctp_send_pending(S0),
    {keep_state, S1};

%% Handle an eNB -> MME data forwarding request (forward)
connected(cast, {send_data, Data}, S0) ->
    sctp_send_from_enb(Data, S0),
    {keep_state, S0};

%% Handle an #sctp_assoc_change event (MME connection state)
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {[], #sctp_assoc_change{state = ConnState,
                                         assoc_id = Aid}}}, S) ->
    case ConnState of
        comm_up ->
            ?LOG_NOTICE("MME connection (id=~p, ~p:~p) is already established?!?",
                        [Aid, MmeAddr, MmePort]),
            {keep_state, S};
        _ ->
            ?LOG_NOTICE("MME connection state: ~p", [ConnState]),
            {stop, {shutdown, conn_fail}}
    end;

%% Handle an #sctp_sndrcvinfo event (MME -> eNB data)
connected(info, {sctp, _Socket, MmeAddr, MmePort,
                 {[#sctp_sndrcvinfo{assoc_id = MmeAid,
                                    stream = SID,
                                    ssn = SSN,
                                    tsn = TSN}], Data}},
          #{mme_aid := MmeAid} = S) ->
    ?LOG_DEBUG("MME connection (id=~p, ~p:~p) -> eNB: ~p",
               [MmeAid, MmeAddr, MmePort,
                #{tsn => TSN, sid => SID, ssn => SSN,
                  len => byte_size(Data), data => Data}]),
    sctp_send_from_mme(Data, S),
    {keep_state, S};

connected(Event, EventData, S) ->
    handle_event(?FUNCTION_NAME, Event, EventData, S).


%% Event handler for all states
handle_event(State, {call, From}, fetch_info, S) ->
    Info = conn_info(State, S),
    {keep_state_and_data, {reply, From, Info}};

handle_event(State, {call, From}, EventData, _S) ->
    ?LOG_ERROR("Unexpected call in state ~p: ~p", [State, EventData]),
    {keep_state_and_data, {reply, From, {error, {unexpected_call, State}}}};

%% Catch-all for unhandled SCTP events
handle_event(State, info, {sctp, _Socket, MmeAddr, MmePort, {AncData, Data}}, _S) ->
    ?LOG_DEBUG("Unhandled SCTP event in state ~p (~p:~p): ~p, ~p",
               [State, MmeAddr, MmePort, AncData, Data]),
    keep_state_and_data;

handle_event(State, info, {sctp_error, _Socket, MmeAddr, MmePort, {_, Error}}, _S) ->
    ?LOG_ERROR("Rx SCTP error in state ~p (~p:~p): ~p",
               [State, MmeAddr, MmePort, Error]),
    sctp_common:report_error(Error),
    keep_state_and_data;

handle_event(State, Event, EventData, _S) ->
    ?LOG_ERROR("Unexpected event ~p in state ~p: ~p", [Event, State, EventData]),
    keep_state_and_data.


code_change(_Vsn, State, S, _Extra) ->
    {ok, State, S}.


terminate(Reason, State,
          #{handler := Pid,
            enb_handle := Handle} = S) ->
    ?LOG_NOTICE("Terminating in state ~p, reason ~p", [State, Reason]),
    enb_registry:enb_unregister(Handle),
    s1ap_proxy:shutdown(Pid),
    case S of
        #{sock := Sock,
          mme_aid := Aid} ->
            sctp_common:shutdown({Sock, Aid}),
            gen_sctp:close(Sock);
        #{sock := Sock} ->
            gen_sctp:close(Sock);
        _ -> ok
    end.


%% ------------------------------------------------------------------
%% private API
%% ------------------------------------------------------------------

%% XXX: proper type
-type state() :: map().


%% Send a single message: eNB -> MME
-spec sctp_send_from_enb(binary(), state()) -> ok | {error, term()}.
sctp_send_from_enb(Data,
                   #{sock := Sock,
                     enb_aid := EnbAid,
                     mme_aid := MmeAid,
                     handler := Pid}) ->
    case s1ap_proxy:process_pdu(Pid, Data) of
        {forward, FwdData} ->
            sctp_common:send_data({Sock, MmeAid}, FwdData);
        {reply, ReData} ->
            sctp_server:send_data(EnbAid, ReData);
        {drop, Data} ->
            ok %% no-op
    end.


%% Send a single message: eNB <- MME
-spec sctp_send_from_mme(binary(), state()) -> ok | {error, term()}.
sctp_send_from_mme(Data,
                   #{sock := Sock,
                     enb_aid := EnbAid,
                     mme_aid := MmeAid,
                     handler := Pid}) ->
    case s1ap_proxy:process_pdu(Pid, Data) of
        {forward, FwdData} ->
            sctp_server:send_data(EnbAid, FwdData);
        {reply, ReData} ->
            sctp_common:send_data({Sock, MmeAid}, ReData);
        {drop, Data} ->
            ok %% no-op
    end.


%% Send pending messages to the MME
sctp_send_pending(#{tx_queue := Pending} = S) ->
    sctp_send_pending(lists:reverse(Pending), S).

sctp_send_pending([Data | Pending], S) ->
    sctp_send_from_enb(Data, S),
    s1gw_metrics:gauge_dec(?S1GW_GAUGE_S1AP_PROXY_UPLINK_PACKETS_QUEUED),
    sctp_send_pending(Pending, S);

sctp_send_pending([], S) ->
    S#{tx_queue := []}.


conn_info(State, #{mme_conn_cfg := MmeConnCfg} = S0) ->
    S1 = maps:with([enb_aid, mme_aid, handler], S0),
    S1#{state => State,
        mme_addr => maps:get(raddr, MmeConnCfg),
        mme_port => maps:get(rport, MmeConnCfg)}.


%% vim:set ts=4 sw=4 et:
